local stuff = require("stuff/lazy")().stuff
local fun = stuff.fun

local matchPrefixInt = function (s, prefix)
	local p = s:sub(1,#prefix)
	if p == prefix then
		local i = tonumber(s:sub(#prefix+1))
		if i > 0 and i == math.floor(i) then
			return i
		end
	end
end

local la
la = stuff.LazyTable(function(k)
	local v = rawget(la, k)
	if v then
		return v
	end
	for _, pf in ipairs({"vec", "mat"}) do
		local num = matchPrefixInt(k, pf)
		if num then
			return la[pf][num]
		end
	end
end)

local makeVec = function(N)
	local mtmt = {}
	local mt = setmetatable({}, mtmt)
	
	mt.__index = mt
	mt.__name = "vec" .. tostring(N)
	mt.N = N
	
	function mtmt:__call(...)
		local args = {...}
		local v = setmetatable({}, mt)
		for i = 1, N do
			v[i] = args[i] or 0
		end
		return v
	end
	
	function mt.assign(v, w)
		for i = 1, N do
			v[i] = w[i] or 0
		end
		return v
	end
	
	function mt.setmeta(t)
		return setmetatable(t, mt)
	end
	
	function mt.nassign(v, n)
		for i = 1, N do
			v[i] = n or 0
		end
		return v
	end
	
	function mt.push(v, x)
		table.insert(v, x)
		return setmetatable(v, la.vec[mt.N + 1])
	end
	
	function mt.pop(v)
		local tmp = table.remove(v)
		return setmetatable(v, la.vec[mt.N - 1])
	end
	
	function mt.assign(v, w)
		for i = 1, N do
			v[i] = w[i]
		end
		return v
	end
	
	function mt.normalize(v)
		local l2 = v:l2()
		if l2 == 0 then
			l2 = 1
		end
		return v:anmul(1/l2)
	end
	
	function mt:copy()
		return setmetatable({unpack(self)}, mt)
	end
	
	local vmap = function(res, f, ...)
		for i = 1, N do
			res[i] = f(i, ...)
		end
		return res
	end
	
	local opMap2 = function(op)
		return function(i, a, b)
			return op(a[i], b[i])
		end
	end
	
	local opMap2L = function(op)
		return function(i, a, b)
			return op(a[i], b)
		end
	end
	
	local opMap1 = function(op)
		return function(i, a)
			return op(a[i])
		end
	end
	
	function mt.amap(v, f)
		for i = 1, N do
			v[i] = f(v[i])
		end
		return v
	end
	
	function mt.map(v, f, out)
		local out = out or setmetatable({}, mt)
		for i = 1, N do
			out[i] = f(v[i])
		end
		return out
	end
	
	for name, op in pairs(fun.operators.dual) do
		local map2 = opMap2(op)
		local map1 = opMap2L(op)
		
		local tf = {true, false}
		
		for _, n in ipairs(tf) do
			local om = (n and opMap2L or opMap2)(op)
			for _, a in ipairs(tf) do
				local pf = (a and "a" or "") .. (n and "n" or "")
				local _a = a
				local _om = om
				mt[pf .. name] = function(a, b, c)
					return vmap(
						_a and a or (c or setmetatable({}, mt)),
						_om,
						a, b
					)
				end
			end
		end
		
		mt["__" .. name] = mt[name]
	end
	
	function mt:l2sq()
		local s = 0
		for i = 1, N do
			s = s + self[i] * self[i]
		end
		return s
	end
	
	function mt:l1()
		local s = 0
		for i = 1, N do
			s = s + math.abs(self[i])
		end
		return s
	end
	
	function mt:l2()
		return math.sqrt(self:l2sq())
	end
	
	function mt:dot(b)
		local s = 0
		for i = 1, N do
			s = s + self[i] * b[i]
		end
		return s
	end
	
	function mt:__tostring()
		local buf = {self.__name or ("vec"..tostring(N)), "("}
		for i = 1, N do
			table.insert(buf, tostring(self[i]))
			if i < N then
				table.insert(buf, ", ")
			end
		end
		table.insert(buf, ")")
		return table.concat(buf)
	end
	
	if N == 2 then
		function mt.ratio(v)
			return v[1] / v[2]
		end
		
		function mt.rotCW(v)
			v[1], v[2] = v[2], -v[1]
			return v
		end
		
		function mt.rotCCW(v)
			v[1], v[2] = -v[2], v[1]
			return v
		end
		
		function mt.fromAngle(a)
			return setmetatable({math.cos(a), math.sin(a)}, mt)
		end
		
		function mt.angle(v)			
			if v[1] == 0 and v[2] == 0 then
				return 0
			else
				return math.atan2(v[2], v[1])
			end
		end
	end
	
	if N == 3 then
		function mt.cross(v1, v2, res)
			local res = res or setmetatable({}, mt)
			for i = 0, 2 do
				res[i+1] = v1[(i+1)%3+1] * v2[(i+2)%3+1] - v1[(i+2)%3+1] * v2[(i+1)%3+1]
			end
			return res
		end
		function mt.v4(v, w, res)
			local res = res or setmetatable({}, la.vec[4])
			for i = 1, 3 do
				res[i] = v[i]
			end
			res[4] = w or 1
			return res
		end
	end
	
	return mt
end

local makeMat = function (N)
	local mt = makeVec(N*N)
	
	mt.__name = "mat" .. tostring(N)
	
	local toIndex0 = function (r, c)
		return r*N + c + 1
	end
	
	local toIndex1 = function (r, c)
		return (r-1)*N + c
	end
	
	function mt.get(m, r, c)
		return m[toIndex1(r,c)]
	end
	
	function mt.set(m, r, c, v)
		m[toIndex1(r, c)] = v
		return m
	end
	
	mt.M = N
	
	local dot3 = function(a, b, res)
		for r = 0, N-1 do
			for c = 0, N-1 do
				local s = 0
				for i = 0, N-1 do
					s = s + a[toIndex0(r, i)] * b[toIndex0(i, c)]
				end
				res[toIndex0(r, c)] = s
			end
		end
		return res
	end
	
	function mt.dot(a, b)
		return dot3(a, b, setmetatable({}, mt))
	end
	
	function mt.adot(a, b)
		local tmp = {}
		for r = 0, N-1 do
			for c = 0, N-1 do
				local s = 0
				for i = 0, N-1 do
					s = s + a[toIndex0(r, i)] * b[toIndex0(i, c)]
				end
				tmp[c+1] = s
				-- c[toIndex0(r, c)] = s
			end
			local off = N*r
			for c = 1, N do
				a[off+c] = tmp[c]
			end
		end
		return c
	end
	
	mt.dot3 = dot3
	
	local vdot3 = function(a, b, c)
		for r = 1, N do
			local s = 0
			for c = 1, N do
				s = s + a[toIndex1(r, c)] * b[c]
			end
			c[r] = s
		end
		return c
	end
	
	mt.vdot3 = vdot3
	
	function mt.vdot(m, v, res)
		return vdot3(m, v, res or setmetatable({}, la.vec[N]))
	end
	
	function mt.transform(m, v, res)
		assert(M == v.N + 1)
		local _ = res == nil or res.N == M - 1 or setmetatable(res, la.vec[M - 1])
		v:push(1)
		local res = mt.vdot(m, v, res)
		v:pop()
		return res, table.remove(res)
	end
	
	local transpose2 = function(m, res)
		for r = 1, N do
			for c = 1, r-1 do
				res[toIndex1(r, c)], res[toIndex1(c, r)]
					= m[toIndex1(c, r)], m[toIndex1(r, c)]
			end
		end
		for i = 1, N do
			res[toIndex1(i,i)] = m[toIndex1(i,i)]
		end
		return res
	end
	
	function mt.transpose(m)
		return transpose2(m, setmetatable({}, mt))
	end
	
	function mt.atranspose(m)
		return transpose2(m, m)
	end
	
	function mt.identity(m)
		local m = m or mt()
		for i = 1, N do
			m[i] = 0
		end
		for i = 0, N-1 do
			m[i*(N+1)+1] = 1
		end
		return m
	end
	
	function mt.getRow(m, r, v)
		local v = v or setmetatable({}, la.vec[N])
		local off = (r-1)*N
		for c = 1, N do
			v[c] = m[off+c]
		end
		return v
	end
	
	function mt.setRow(m, r, v)
		local off = (r-1)*N
		for c = 1, N do
			m[off+c] = v[c]
		end
		return m
	end
	
	function mt.getCol(m, c, v)
		local v = v or setmetatable({}, la.vec[N])
		for r = 0, N-1 do
			v[r+1] = m[c+r*N]
		end
		return v
	end
	
	function mt.setCol(m, c, v)
		for r = 0, N-1 do
			m[c+r*N] = v[r+1]
		end
		return m
	end
	
	function mt.setCorner(m, m2)
		local minN = math.min(m.M, m2.M)
		for r = 1, minN do
		for c = 1, minN do
			m:set(r, c, m2:get(r, c))
		end end
		return m
	end
	
	function mt.getCorner(m, n, res, diag)
		local res = res or la.mat[n]()
		local minN = math.min(n, N)
		for i = minN+1, n do
			res:set(i,i,diag)
		end
		for r = 1, minN do
		for c = 1, minN do
			res:set(r, c, m:get(r, c))
		end end
		return res
	end
	
	function mt.toGaussRows(...)
		local mats = {...}
		local vec = la.vec[#mats*N]
		local rows = {}
		for r = 1, N do
			local tmp = {}
			for i = 1, #mats do
				for c = 1, N do
					tmp[(i-1)*N+c] = mats[i][(r-1)*N+c]
				end
			end
			table.insert(rows, setmetatable(tmp, vec))
		end
		return rows
	end
	
	mt.__concat = mt.dot
	
	function mt.gauss(rows)
		for r = 1, N do
			local bestValue = 0
			local bestRow
			for r2 = r, N do
				local x = math.abs(rows[r2][r])
				if x >= bestValue then
					bestRow = r2
					bestValue = x
				end
			end
			if bestValue == 0 then
				return nil
			end
			
			rows[r],rows[bestRow] = rows[bestRow],rows[r]
			rows[r]:anmul(1.0/rows[r][r])
			
			local vec = getmetatable(rows[1])
			local tmp = vec()
			for r2 = 1, r-1 do
				rows[r]:nmul(rows[r2][r], tmp)
				rows[r2]:asub(tmp)
			end
			for r2 = r+1, N do
				--~ print(rows[r], rows[r2][r], tmp)
				rows[r]:nmul(rows[r2][r], tmp)
				rows[r2]:asub(tmp)
			end
		end
		return rows
	end
	
	function mt.afromGauss(m, rows, i)
		for r = 1, N do
		for c = 1, N do
			m[toIndex1(r, c)] = rows[r][c+(i-1)*N]
		end end
		return m
	end
	
	function mt.fromGauss(rows, i)
		return mt.afromGauss(setmetatable({}, mt), rows, i)
	end
	
	local identity = mt.identity()
	
	function mt.ainvert(m, m2)
		return mt.afromGauss(m, mt.gauss(mt.toGaussRows(m2, identity)), 2)
	end
	
	function mt.invert(m)
		return mt.ainvert(setmetatable({}, mt), m)
	end
	
	function mt:__tostring()
		local buf = {self.__name or ("mat"..tostring(N)), "("}
		for i = 1, N*N do
			if (i - 1) % N == 0 then
				table.insert(buf, "\n\t")
			end
			table.insert(buf, tostring(self[i]))
			table.insert(buf, ",\t")
		end
		table.insert(buf, "\n)")
		return table.concat(buf)
	end
	
	if N == 2 then
		function mt.rotate(angle)
			local c = math.cos(angle)
			local s = math.sin(angle)
			return setmetatable({
				c, s,
				-s, c,
			}, mt)
		end
	end
	
	if N == 4 then
		function mt.rotateAxis(axis, theta)		
			local m = mt.identity()
			local cos = math.cos(theta)
			local sin = math.sin(theta)
			local i1 = (axis) % 3 + 1;
			local i2 = (axis + 1) % 3 + 1;
			m:set(i1, i1, cos)
			m:set(i2, i1, -sin)
			m:set(i1, i2, sin)
			m:set(i2, i2, cos)
			return m
		end
		
		function mt.translate(v)
			local m = mt.identity()
			for i = 1, v.N or #v do
				m:set(i, N, v[i])
			end
			return m
		end
		
		--~ function mt.perspective(div, near, far)
			--~ return la.mat4(
				--~ 1,	0,	0,		0,
				--~ 0,	1,	0,		0,
				--~ 0,	0,	1,		0,
				--~ 0,	0,	div,	0
			--~ )
		--~ end
		
		function mt.perspective(aspect, fov, near, far)
			local a = -math.tan(fov/2.0)
			local b = a*(far+near)/(near-far)
			local c = near*b - near*a
			return la.mat4(
				1,	0,		0,	0,
				0,	aspect,	0,	0,
				0,	0, 		-b,	-c,
				0,	0,		a,	0
			)
		end
		
		function mt.perspective2(aspect_ratio, fov_ratio, near, far)
			local a = 1 / fov_ratio
			local b = (near + far) / (far - near)
			--~ local d = far * (1 - b)
			local d = far * (b - 1) -- far is in z- direction
			return la.mat4(
				a, 0, 0, 0,
				0, a*aspect_ratio, 0, 0,
				0, 0, -b, -d,
				0, 0, -1, 0
			)
		end
		
		function mt.scale(v)
			local m = mt.identity()
			for i = 1, v.N or #v do
				m:set(i,i,v[i])
			end
			return m
		end
		
		function mt.nscale(n)
			local m = mt.identity()
			for i = 1, 3 do
				m:set(i,i,n)
			end
			return m
		end
	end
	
	return mt
end

la.vec = stuff.LazyTable(makeVec)
la.mat = stuff.LazyTable(makeMat)

return la
